# -*- coding: utf-8 -*-

'''
Model

Manages everything with paths (from config), QGIS environment,
working on file system (create, delete).
For the special works, such as transformations, own classes exists.

Created on 23.11.2017
last change: 08.11.2019
@author: Weatherman
'''

import os
from os import path
from pathlib import Path
import sys

from configparser import ConfigParser    # configparser in Python3
from glob import glob
import platform    # to determine wether Windows or not

from qgis.core import QgsApplication, QgsProject
# For the QGIS printing template:
from qgis.PyQt.QtXml import QDomDocument

import product_definitions    # File 'product_definitions.py'
CONFIG_NAME = "config.ini"



class Model:
    '''
    classdocs
    '''
    
    
    def __init__(self):
        '''
        Constructor
        '''
        
        print(self)
        
        self._plugin_dir = path.abspath(path.join(path.dirname(__file__), os.pardir))
        
        config_file = path.join(self._plugin_dir, CONFIG_NAME)
        
        if not path.exists(config_file):
            raise FileNotFoundError("Config file '{}'".format(config_file))
        
        #
        # Parameter
        #
        self._config_file = config_file
        
        #
        # Ermittelt
        #
        
        # Objekt, auf das wir im Folgenden immer zugreifen.
        # Statisch ging es leider aus welchen Gründen auch immer (war in anderen Methoden 'None').
        self._config = ConfigParser()
        self._config.read(config_file)
        
        self._product_defs = product_definitions.dict_titles
        self.out("{}: {} product titles loaded.".format(product_definitions.__name__, len(self._product_defs)))
        
        self._data_dir   = None
        self._temp_dir   = None
        
        self._list_of_used_files = None    # []
        
        
        
    
    def __str__(self):
        return self.__class__.__name__
    
    
    def out(self, s, ok=True):
        if ok:
            print("{}: {}".format(self, s))
        else:
            print("{}: {}".format(self, s), file=sys.stderr)
    
    
    
    def create_data_storage_and_clean(self):
        self.out("create_data_storage_and_clean()")
        
        temp_dir = self.temp_dir    # important NOT to use '_temp_dir' here!
        
        try:
            os.makedirs(temp_dir)    # inklusive data dir
        except FileExistsError:
            # Remove old content:
            print("  remove temp dir contents...")
            # If ignore_errors is set, errors are ignored; otherwise, if onerror is set, it is called to handle ...
            #shutil.rmtree(temp_dir)
            #shutil.rmtree(path, ignore_errors=True)
            # -> Ganzes Verzeichnis löschen ist zu problematisch ("in use").
            for f in glob( path.join(temp_dir, "*") ):
                try:
                    os.remove(f)
                # Merkwürdigerweise trat eine Exception unter Windows (7) auf, wenn man mehrmals
                # Daten lädt. Irgendwie lässt das System die angefassten Daten nicht los.
                except WindowsError as e:    # only available on Windows
                    self.out("ERROR: {}\n  versuche zu ignorieren.".format(e))
            # for
        else:
            self.out("temp dir created: {}".format(temp_dir))
        
    
    
    def handle_history_file_content(self, new_file):
        """
        hint: 'l_files' was filled in method "_init_dock()"
        """
        
        # File exists - so it can be already stored for later suggestion for the user:
        
        # write completely new:
        if not self.list_of_used_files:
            with open(self.history_file, 'w') as hf:
                hf.write(new_file + '\n')
            return    # finish
        
        # handle existing content
        
        # Try to remove _this_ file (if contains), so it can be new positioned at the top:
        try:
            self.list_of_used_files.remove(new_file)
        except ValueError:
            pass
        
        self.list_of_used_files.insert(0, new_file)
        
        _max = 9
        # write the _max items back:
        with open(self.history_file, 'w') as hf:
            for i, line in enumerate(self.list_of_used_files):
                if i >= _max:
                    break
                hf.write(line + '\n')
            
    
    
    def create_qdocument_from_print_template_content(self):
        print_template = self.default_print_template
        
        self.out("create_qdocument_from_print_template_content(): {}".format(print_template))
        
        if not print_template:
            raise FileNotFoundError()
        
        # Load template
        with open(print_template, 'r') as templateFile:
            print_template_content = templateFile.read()
        
        q_xmldoc = QDomDocument()
        # If namespaceProcessing is true, the parser recognizes namespaces in the XML file and sets
        # the prefix name, local name and namespace URI to appropriate values. If namespaceProcessing
        # is false, the parser does no namespace processing when it reads the XML file.
        q_xmldoc.setContent(print_template_content, False)    # , bool namespaceProcessing
        
        return q_xmldoc
        
    
    def product_description(self, prod_id):
        try:
            return self._product_defs[prod_id]
        except KeyError:
            return "Product has not been defined yet."
    
    
    def title(self, dt_date, prod_id):
        ''' Zugriff auf Title im Dictionary. '''
        
        s_date = dt_date.strftime("%d.%m.%Y, %H:%M UTC")
        
        return "{} des {}".format(self.product_description(prod_id), s_date)
    
    
    def qml_file(self, precision, prod_id):
        '''
        precision: 1.0, 0.1, 0.01
        prod_id:   Product-ID for determining the QML-File
        '''
        
        """
        Gemeinsame Farbskalen:
        Hierfür werden jetzt die Produkt-Summenzeiträume definiert.
        Anordnung: von klein nach groß
        """
        #min_products     = ('RZ', 'RY', 'YW', 'EZ', 'EY')    # 5 minute products
        #hourly_products  = ('RW', 'S2', 'S3')
        daily_products   = ('SF', 'D2', 'D3')
        monthly_products = ('W1', 'W2', 'W3', 'W4', 'SM')
        yearly_products  = ('SJ', 'SY', 'JW')
        
        qml_fn = 'hourly'
        
        if precision < 0.1:    # <- incl. RX RVP6 to mm conversion -> precision changed to 0.01
            qml_fn = '5min'
        elif prod_id in daily_products:
            if prod_id.startswith('D'):
                qml_fn = 'daily+'    # eine weitere Abstufung
            else:
                qml_fn = 'daily'
        elif prod_id in monthly_products:
            qml_fn = 'monthly'
        elif prod_id in yearly_products:
            qml_fn = 'yearly'
        elif prod_id[1] == 'X':    # products in RVP6-Units: RX, WX, EX
            qml_fn = 'rvp6units'
        else:
            self.out("qml_file(): using default")
        
        qml_file = path.join(self.symbology_path,  qml_fn+".qml")
        
        # QML file exists?
        if not path.exists(qml_file):
            self.out("qml_file(): QML file '{}' not found.".format(qml_file), False)
            qml_file = None
        
        return qml_file
    
    
    """
    Properties
    """
    
    # Dirs:
    
    @property
    def profile_dir(self):
        return QgsApplication.qgisSettingsDirPath()
    
    @property
    def plugin_dir(self):
        return self._plugin_dir
    @property
    def data_dir(self):
        ''' read from definition file
        raise FileNotFoundError '''
        if self._data_dir:    # cached?
            return self._data_dir
        with open(self.data_dir_def_file) as f:    # FileNotFoundError
            self._data_dir = f.read()
        return self._data_dir
    @data_dir.setter
    def data_dir(self, d):
        ''' update, when user writes a new path in data dir def file '''
        self.out("update 'data_dir': {}".format(d))
        self._data_dir = d
    @property
    def temp_dir(self):
        return path.join(self.data_dir, 'tmp')    # important NOT to use '_data_dir' here!
    @property
    def layout_dir(self):
        return path.join(self._plugin_dir, self._config.get('Paths', 'LAYOUT_PATH'))
    
    @property
    def list_of_used_files(self):
        return self._list_of_used_files
    @list_of_used_files.setter
    def list_of_used_files(self, l):
        self._list_of_used_files = l
    
    @property
    def default_border_shape(self):
        """ gleich mit Check """
        #border_shape = path.join(self._plugin_dir, self._config.get('Paths', 'CUT_TO'))
        #
        #if not path.exists(border_shape):
        #    self.out("default shape mask '{}' not found!".format(border_shape), False)
        #    return None
        
        border_shape = Path(self._plugin_dir) / self._config.get('Paths', 'CUT_TO')
        
        if not border_shape.exists():
            self.out("default shape mask '{}' not found!".format(border_shape), False)
            return None
        
        return str(border_shape)    # convert as string, otherwise error "Posix path"
    
    @property
    def default_print_template(self):
        # verschiedene Namen wegen internem Pfad für Logo:
        fn = "Druckvorlage.qpt" if not platform.system() == 'Windows'  else "Druckvorlage_win.qpt"
        
        #print_template = path.join( self._config.get('Paths', 'RESSOURCE_PATH'), "Layout/{}".format(fn) )
        print_template = Path(self._config.get('Paths', 'RESSOURCE_PATH')) / "Layout" / fn
        
        #if not path.exists(print_template):
        if not print_template.exists():
            self.out("default print template '{}' not found!".format(print_template), False)
            return None
        
        return print_template
    
    @property
    def default_project(self):
        #project_file = self._config.get('Paths', 'DEFAULT_PROJECT')
        #project_file = path.join(self._plugin_dir, project_file)
        project_file = Path(self._plugin_dir) / self._config.get('Paths', 'DEFAULT_PROJECT')
        
        #if not path.exists(project_file):
        if not project_file.exists():
            self.out("default project '{}' not found!".format(project_file), False)
            return None
        
        return str(project_file)
    
    
    @property
    def symbology_path(self):
        #symb_path = self._config.get('Paths', 'STYLE_PATH')
        #symb_path = path.join(self._plugin_dir, symb_path)
        symb_path = Path(self._plugin_dir) / self._config.get('Paths', 'STYLE_PATH')
        
        #if not path.exists(symb_path):
        if not symb_path.exists():
            self.out("symbology path '{}' not found!".format(symb_path), False)
            return None
        
        return str(symb_path)
    
    @property
    def data_dir_def_file(self):
        file = Path(self.profile_dir) / self._config.get('Paths', 'datadir_deffile_basename')
        return str(file)
    @property
    def history_file(self):
        return Path(self.profile_dir) / self._config.get('Paths', 'last_products_basename')
    @property
    def check_file(self):
        return Path(self.plugin_dir) / '__check__'
    
    
    @property
    def projection_radolan(self):
        return self._config.get('Projections', 'PROJ_RADOLAN_NATIVE')
    
    @property
    def projection_laea(self):
        return self._config.get('Projections', 'PROJ_EPSG3035_DEST')
    

