# -*- coding: utf-8 -*-
"""
RADOLAN to map (radolan2map)

A QGIS plugin to bring a RADOLAN binary file onto a map.

changed: 2019-09
email  : radolan2map@e.mail.de

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os
from os import path       # shortcut
from pathlib import Path
from glob import glob
import sys
import platform    # to determine wether Windows or not
import time

# Import the PyQt and QGIS libraries
from qgis.PyQt.QtCore import Qt, QCoreApplication, QSettings, QFileInfo, QTranslator, QSize, pyqtSignal #, QRect
from qgis.PyQt.QtWidgets import QMessageBox, QDockWidget, QAction, QFileDialog, QDialogButtonBox #, QDialog
from qgis.PyQt.QtGui import QIcon

from qgis.core import Qgis, QgsProject, QgsRasterLayer, QgsMapLayer, QgsLayoutManager, QgsPrintLayout
from qgis.core import QgsReadWriteContext, QgsLayoutItemLegend, QgsLayerTree, QgsLayerTreeLayer, QgsRasterTransparency

# Initialize Qt resources from file resources.py
#from .resources import *

from console import console    # Python-Konsole automatisch zeigen

# Eigene Klassen
# .........................................................
class_path = path.join(path.dirname(__file__), 'classes')
sys.path.append(class_path)    # Verzeichnis der weiteren Klassen
from gui import DockWidget
from Model import Model
from RadolanBin2AsciiConverter import RadolanBin2AsciiConverter
from RadolanBin2AsciiConverter import test_product_get_id    # import a function
from GDALProcessing import GDALProcessing
# .........................................................



class Radolan2Map:
    """ QGIS Plugin Implementation. """
    
    def __init__(self, iface):
        """Constructor.
        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        
        # Save reference to the QGIS interface
        self.iface = iface
        
        
        # initialize plugin directory
        self.plugin_dir = path.dirname(__file__)
        
        # initialize locale
        # Bugfix: on Windows 10 and QGIS 3.6 Noosa occured a TypeError.
        # The reason could be, that there are no translation files.
        try:
            locale = QSettings().value('locale/userLocale')[0:2]
        except TypeError as e:
            self.out(e, False)    # TypeError: 'QVariant' object is not subscriptable
        else:
            locale_path = path.join(
                self.plugin_dir,
                'i18n',
                'Radolan2Map_{}.qm'.format(locale))
            
            if path.exists(locale_path):
                self.translator = QTranslator()
                self.translator.load(locale_path)
                QCoreApplication.installTranslator(self.translator)
        
        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&radolan2map')
        # TODO: We are going to let the user set this up in a future iteration
        self.toolbar = self.iface.addToolBar(u'Radolan2Map')
        self.toolbar.setObjectName(u'Radolan2Map')

        
        self._model = Model()
        
        # RADOLAN-Converter bin -> asc:
        self._ascii_converter = None
        
        # Check if plugin was started the first time in current QGIS session
        # Must be set in initGui() to survive plugin reloads
        # Create the dialog (after translation) and keep reference
        self.dock = None
        
        
        
        
    
    
    def __str__(self):
        return self.__class__.__name__
    
    
    def out(self, s, ok=True):
        if ok:
            print("{}: {}".format(self, s))
        else:
            print("{}: {}".format(self, s), file=sys.stderr)
    
    
    
    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('Radolan2Map', message)
    
    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar. """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToRasterMenu(
                self.menu,
                action)

        self.actions.append(action)
        
        return action
    
    
    
    def initGui(self):
        """ Create the menu entries and toolbar icons inside the QGIS GUI. """
        
        """
        note the icon path:
        ':/plugins/my_plugin/icon.png'.
        The colon instructs QGIS to use the compiled resources.py file to locate the icon.
        Open the original resources.qrc file and you'll see the connection.
        """
        #icon_path = ':/plugins/radolan2map/icon.png'
        icon_path = path.join(path.dirname(__file__), "img/icon.png")
        
        self.add_action(
            icon_path,
            text=self.tr(u'radolan2map: load a RADOLAN binary file'),    # tooltip over plugin icon
            callback=self.open_dock,
            parent=self.iface.mainWindow())

    
    def onClosePlugin(self):
        """ Cleanup necessary items here when plugin dockwidget is closed """
        
        #print "** CLOSING DockPlugin"
        
        # disconnects
        self.dock.closingPlugin.disconnect(self.onClosePlugin)

        # remove this statement if dockwidget is to remain
        # for reuse if plugin is reopened
        # Commented next statement since it causes QGIS crashe
        # when closing the docked window:
        # self.dock = None
    
    
    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""

        #print "** UNLOAD Radolan2Map"
        
        for action in self.actions:
            self.iface.removePluginRasterMenu(
                self.tr(u'&radolan2map'),
                action)
            self.iface.removeToolBarIcon(action)
        # remove the toolbar
        del self.toolbar
    
    
    
    def select_input_file(self):
        text_input_field = self.dock.cbbox_radolan.currentText()
        
        # Determine start directory:
        # We assume a set up file here.
        if text_input_field:
            file = text_input_field
        else:
            file = QgsProject.instance().fileName()    # project file
        
        start_dir = path.dirname(file)
        
        """
        raa01-rw_10000-1708020250-dwd---bin
        raa01-rw_10000-1708020250-dwd---bin.gz
        adjust.pix """
        # You can also add other filter. You need to separate them with a double ;; like so :
        # "Images (*.png *.xpm .jpg);;Text files (.txt);;XML files (*.xml)"
        file_filter = "RADOLAN binary files (*bin* *.pix);;Unknown product format (*)"
        
        input_file, _ = QFileDialog.getOpenFileName(self.dock, "Input binary RADOLAN file", start_dir, file_filter,
                        # these additional parameters are used, because QFileDialog otherwise doesn't start with the given path:
                        None, QFileDialog.DontUseNativeDialog)
        
        if not input_file:
            return    # preserve evtl. filled line
        
        # Set:
        #self.dock.inputpath.setText(input_file)
        combo = self.dock.cbbox_radolan    # shorten
        
        """ Set selection to first (just inserted) position.
        This triggers the change event listener which is then executed twice.
        So we block the emitting of a signal during index change: """
        combo.blockSignals(True)
        combo.insertItem(0, input_file)    #combo.addItem(input_file)    # addItem simplest method
        combo.blockSignals(False)
        combo.setCurrentIndex(0)    # trigger if there is a real selection change (more than one element)
        
        # If one element now, the event listener was not triggered,
        # because there is no selection change. We trigger manually:
        if combo.count() == 1:
            self._combo_selection_change()
        
        
    
    def _combo_selection_change(self):
        #file = self.dock.inputpath.text()
        file = self.dock.cbbox_radolan.currentText()
        self.out("selection changed: '{}'".format(file))
        
        #
        # Check input file:
        #
        
        action_btn = self.dock.btn_action    # shorten
        
        if not file:
            print("  empty entry ignored")
            action_btn.setEnabled(False)
            return
        
        if not Path(file).exists():
            QMessageBox.critical(self.iface.mainWindow(),
                'File error', "The specified file doesn't exist!")
            self._model.list_of_used_files.remove(file)
            self.out("file '{}' removed from list".format(file))
            action_btn.setEnabled(False)
            return
        
        
        # Test product, if it is a 'X'-product (RX, WX, EX) with values coded as RVP6-units:
        prod_id = test_product_get_id(file)
        
        is_rx = True if prod_id[1] == 'X'  else False
        """ Further string tests on product id don't make sense,
        because IDs like 'W1'-'W4', 'D2'-'D3', 'S2'-'S3' are possible. """
        
        if is_rx:
            self.out("*X product detected - '{}' checkbox enabled".format(self.dock.check_rvp6tomm.objectName()))
        self.dock.check_rvp6tomm.setVisible(is_rx)
        
        # Print info found in def-file about the product:
        prod_desc = self._model.product_description(prod_id)
        self.out('product {}: "{}"'.format(prod_id, prod_desc))
        
        #self.dock.btn_show_list.setVisible(True)
        # after selection of RADOLAN file enable action button:
        action_btn.setEnabled(True)
        
        #self._load_available_files_to_list(path.dirname(file))
    
    
    '''
    def _load_available_files_to_list(self, folder):
        files = []
        for p in ("*bin*", "*.pix"):    # compare with above
            pattern = path.join(folder, p)
            names = [path.basename(x) for x in glob(pattern)]
            files.extend(names)
        
        if files:
            self.dock.list_widget.add_items(files)
    '''
    
    
    def select_symbology(self):
        file_filter = "QGIS symbology files (*.qml*)"
        
        inputqmlstyle, _ = QFileDialog.getOpenFileName(self.dock, "Apply symbology (optional)",
                                                       self._model.symbology_path, file_filter,
        # these additional parameters are used, because QFileDialog otherwise doesn't start with the given path:
                                                       None, QFileDialog.DontUseNativeDialog)
        if not inputqmlstyle:
            return    # evtl. geladener Pfad nicht löschen
        self.dock.inputqml.setText(inputqmlstyle)
    
    
    
    def open_dock(self):
        """
        This method will be called when you click the toolbar button or select the plugin menu item.
        """
        
        
        # Ausgaben des Plugins zeigen:
        #console.show_console()    # geht, aber besser:
        pythonConsole = self.iface.mainWindow().findChild( QDockWidget, 'PythonConsole' )
        try:
            if not pythonConsole.isVisible():
                pythonConsole.setVisible(True)
        except AttributeError:    # can be 'None' above
            self.out("PythonConsole Error catched", False)
            console.show_console()
        
        
        
        self.out("open_dock()")
        
        # Create the dialog with elements (after translation) and keep reference
        # Only create GUI ONCE in callback, so that it will only load when the plugin is started
        
        # dockwidget may not exist if:
        #    first run of plugin
        #    removed on close (see self.onClosePlugin method)
        
        
        # Close dock via toolbar button:
        if self.dock:
            if self.dock.isVisible():
                self.dock.close()
                return
        # if first start
        else:
            """
            Detect first start after installation.
            There seems to be a problem with plugin reloading, so after
            (re)install there occur errors. So we handle this issue manually with
            a one time check file (deleted after the following message).
            """
            if self._model.check_file.exists():
                msg = "After reinstalling the plugin\nQGIS should be restarted now,"\
                " otherwise plugin errors are to be expected!"
                QMessageBox.warning(self.iface.mainWindow(), 'initialize plugin', msg)
                # Message should only appear one time, so the
                # check file must removed now:
                self._model.check_file.unlink()
                self.out("check file '{}' removed".format(self._model.check_file))
            # if
            self._init_dock()
        
        
        """
        DockWidget exists but maybe it's invisible
        Bring DockWidget to front (maybe it's minimized elsewhere...)
        """
        width  = self.dock.frameGeometry().width()
        height = self.dock.frameGeometry().height()
        # das ist schon gut - ggf. die init. Werte vom .ui holen
        self.dock.setMinimumSize(QSize(width, height))    # width, height
        # -> verhindert ein zu kleines Widget
        
        # show the dockwidget
        self.iface.addDockWidget(Qt.RightDockWidgetArea, self.dock)
        self.dock.show()
        # Ohne diese Anweisung wurde das Fenster ab QGIS3 im Hintergrund geöffnet.
        #self.dock.setWindowFlags(Qt.WindowStaysOnTopHint)
        
        
        ''' not in dock widget
        # Run the dialog event loop
        #result = self.dlg.exec_()
        
        # See if OK was pressed
        if not result:
            self.out("There is nothing to do.")
            self.dlg.close()
            return
        '''
        
    
    def _init_dock(self):
        # Create the dockwidget (after translation) and keep reference
        self.dock = DockWidget()
        
        # connect to provide cleanup on closing of dockwidget
        self.dock.closingPlugin.connect(self.onClosePlugin)
        
        
        """
        set connections
        """
        
        # Button for loading a prepared template project:
        self.dock.btn_load_project.clicked.connect(self.ask_user_load_template_project)
        # Button für schnellen Test des PrintLayouts nutzen:
        #self.dock.btn_load_project.clicked.connect(self._load_print_layout )    # !: Funktion ohne ()
        self.dock.filedialog_input.clicked.connect(self.select_input_file)
        
        self.dock.filedialog_qml.clicked.connect(self.select_symbology)
        # connect accept and cancel:
        #self.dock.button_box.accepted.connect(self._run)
        # For Apply-Button you'll need to connect the clicked signal manually in your widget:
        #apply_btn = self.dock.button_box.button(QDialogButtonBox.Apply)
        #apply_btn.clicked.connect(self._run)
        self.dock.btn_action.clicked.connect(self._run)
        
        
        """
        resets / defaults
        """
        
        ############################
        # Tab action
        ############################
        
        combo = self.dock.cbbox_radolan    # shorten
        
        # read previously used files and prefill combo box for the user:
        history_file = self._model.history_file
        if history_file.exists():
            self.out("reading 'history file' {}".format(history_file))
            with open(history_file) as hf:
                self._model.list_of_used_files = hf.read().splitlines()    # lines without '\n'
                # hint: 'list_of_used_files' will be used further
            # fill:
            combo.addItem('')    # first a empty item (no valid selection)
            for file in self._model.list_of_used_files:    # first items in the file also positioned first in list
                combo.addItem(file)
        # history file
        
        # Register event handler only here, so it is not triggered at filling above:
        combo.currentIndexChanged.connect(self._combo_selection_change)
        
        
        self.dock.inputmask.setText( self._model.default_border_shape )    # DEU_adm0
        
        
        ############################
        # Tab TIF storage
        ############################
        
        self.dock.btn_save.clicked.connect(self.write_new_storage_location)
        # Try to show the existing data folder, but maybe the file does not exist yet:
        try:
            self.dock.storage_dir = self._model.data_dir
        except FileNotFoundError:
            pass
        
        
        """
        If data def file doesn't exist, the action tab will be disabled
        """
        tab_to_activate_index = 0
        if path.exists(self._model.data_dir_def_file):
            self.out("'data dir def file' found:\n  {}".format(self._model.data_dir_def_file))
        else:
            self.out("'data dir def file' not found -> init")
            #self.dock.tabWidget.removeTab(0)    # -> working with the plugin is not possible
            self.dock.tabWidget.setTabEnabled(tab_to_activate_index, False)    # disable first tab "operation"
            tab_to_activate_index = 2    # tab for define storage folder
        
        self.dock.tabWidget.setCurrentIndex(tab_to_activate_index)
        
    
    
    def _run(self):
        """
        Run method that performs all the real work.
        """
        
        file = self.dock.cbbox_radolan.currentText()
        
        # File exists (checked before) - so it can be already stored for later suggestion for the user:
        self._model.handle_history_file_content(file)
        self.out("selected file stored in '{}'".format(self._model.history_file))
        
        
        clip_to_mask = False
        inputmaskshp = self.dock.inputmask.text()
        # Checkbox enabled AND mask shape specified?
        if self.dock.check_cut.isChecked()  and  inputmaskshp:
            clip_to_mask = True
            sep = '\\' if platform.system() == 'Windows'  else '/'
            inputmaskshp = inputmaskshp.replace(sep, os.sep)
        
        
        # + + + + + + + + + + + + + + + +
        
        
        
        self._model.create_data_storage_and_clean()
        
        
        #exclude_zeroes = self.dock.check_excl_zeroes.isChecked()
        #self._ascii_converter = RadolanBin2AsciiConverter(input_file, self._model.temp_dir, exclude_zeroes)
        # -> now better with layer renderer - conserves 0 values
        self._ascii_converter = RadolanBin2AsciiConverter(file, self._model.temp_dir)
        
        # RX to mm?
        if self.dock.check_rvp6tomm.isVisible()  and  self.dock.check_rvp6tomm.isChecked():
            self._ascii_converter.rvp_to_mm = True
        self._ascii_converter.run()    # ggf. Exception
        
        #prod_id = self._ascii_converter.prod_id
        
        filename, dim, _max, _min, mean, total, valid, nonvalid = self._ascii_converter.statistics
        s_max = str(_max)
        # if part after point is to long:
        l_max = s_max.split('.')
        if len(l_max[1]) > 2:
            s_max = "{:.2f}".format(_max)
        
        self.dock.text_filename.setText(filename)
        self.dock.text_shape.setText(dim)
        self.dock.text_max.setText(s_max)
        self.dock.text_min.setText(str(_min))
        self.dock.text_mean.setText("{:.2f}".format(mean))
        self.dock.text_total_pixels.setText(str(total))
        self.dock.text_valid_pixels.setText(str(valid))
        self.dock.text_nonvalid_pixels.setText(str(nonvalid))
        
        self.dock.tabWidget.setTabEnabled(1, True)    # second tab "statistics"
        
        
        # -> .asc-Name gesetzt
        # + + + + + + + + + + + + + + + +
        
        
        # define name of (clipped) TIFF file (based on .asc filename):
        tif_extension = '_clipped.tif' if clip_to_mask  else '.tif'
        
        full_asc_filename = self._ascii_converter.full_asc_filename
        tif_bn = path.basename(full_asc_filename).replace('.asc', tif_extension)    # tif basename
        full_tif_filename = path.join(self._model.data_dir, tif_bn)
        
        
        gdal_processing = GDALProcessing(self._model, full_asc_filename, full_tif_filename)
        #gdal_processing.produce_warped_tif_using_script()
        
        shapefile = inputmaskshp if clip_to_mask  else None
        # Clip to the given mask (optional):
        gdal_processing.produce_warped_tif_by_python_gdal(shapefile)
        
        
        """
        From here is related to GIS layer loading
        """
        
        raster_to_load = gdal_processing.tif_file
        
        
        result_layer = QgsRasterLayer(raster_to_load, path.basename(raster_to_load) )
        
        # this too? Lead to layer loaded twice.
        #QgsMapLayerRegistry.instance().addMapLayer(result_layer)
        
        if not result_layer.isValid():
            msg = "Layer '{}' failed to load!".format(path.basename(raster_to_load))
            self.out(msg, False )
            self.iface.messageBar().pushMessage("Error", msg, level=Qgis.Critical, duration=3)
            return
        
        """
        Das neue Rasterlayer ist i.O. und es wird im Folgenden hier weiter damit gearbeitet
        """
        
        
        radolan_layer_name = self._ascii_converter.layer_name
        if clip_to_mask:
            radolan_layer_name += "_clipped"
        
        
        """
        --- Clean: Ggf. vorhandenes Raster entfernen ---
        
        Wenn schon ein Rasterlayer mit obigem Namen existiert, wird dieses zuvor entfernt.
        Die "Legende" hat hier wohl nichts mit der Composer-Legende zu tun!
        """
        
        #for layer in self.iface.legendInterface().layers():    # QGIS 2
        #for layer in self.iface.layerTreeView().selectedLayers():
        raster_layers = [rl for rl in QgsProject.instance().mapLayers().values() if rl.type() == 1]
        self.out("current project: {} raster layers found.".format(len(raster_layers)))
        
        for layer in raster_layers:
            #if layer.type() == QgsMapLayer.RasterLayer  and  layer.name() == radolan_layer_name:
            if layer.name() == radolan_layer_name:
                self.out('RasterLayer with existing name "{}" found.'.format(layer.name()))
                """ Ausgabe
                1 Raster
                0 DEU_adm0
                """
                #print(layerType, layer.name())
                
                print("  removing layer \"{}\"".format(layer.name()))
                QgsProject.instance().removeMapLayer( layer.id() )
        # for
        
        
        #
        # Load result raster
        #
        
        # Vorherige Version ohne Insert an bestimmter Position (ging auch):
        #self.iface.addRasterLayer(raster_to_load, path.basename(raster_to_load) )
        
        # Einfaches result_layer.setLayerName('Raster') bringt leider nichts.
        # Unser Rasterlayer kann zuverlässig so bestimmt werden (kann gleich noch einmal gebraucht werden):
        #active_raster_layer = self.iface.activeLayer()
        active_raster_layer = result_layer
        active_raster_layer.setName(radolan_layer_name)
        
        
        
        """
        If no project file not loaded when running plugin
        """
        
        project = QgsProject.instance()
        # Print the current project file name (might be empty in case no projects have been loaded)
        #self.out("bisheriges Projekt: {}".format(project.fileName()))
        
        write = False
        
        if not path.exists(project.fileName()):    # prepare project
            self.out("### no project open -> create a new one ###")
            project = self.load_project()
            write = True
        
        
        # Insert layer at a certain position
        
        # Add the layer to the QGIS Map Layer Registry (the second argument must be set to False
        # to specify a custom position:
        project.addMapLayer(active_raster_layer, False)  # first add the layer without showing it
        
        # obtain the layer tree of the top-level group in the project
        layerTree = self.iface.layerTreeCanvasBridge().rootGroup()
        # The position is a number starting from 0, with -1 an alias for the end.
        # Index 0: ganz oben, Index 1: 2. Position unter der Vektor-Layer-Gruppe:
        layerTree.insertChildNode(1, QgsLayerTreeLayer(active_raster_layer))
        
        
        
        # Set symbology from QML file:
        # 1) as given parameter by user or
        # 2) automatically by program
        qml_file = None
        
        # self defined symbology:
        if self.dock.check_symb.isChecked():
            qml_file = self.dock.inputqml.text()
            if not qml_file:    # empty field (string)?
                self.out("QML field = emtpy - choose one automatically", False)
            
        # Determine prepared QML-File delivered with the plugin:
        if not qml_file:
            qml_file = self._model.qml_file(self._ascii_converter.precision, self._ascii_converter.prod_id)
            # <prod_id>.qml or 'None'
        
        if qml_file:
            self.out("using QML file '{}'".format(qml_file))
            
            # merkwürdig: das geht nicht:
            #result_layer.loadNamedStyle(qml_file)
            
            #if layer.geometryType() == QGis.Point:
            active_raster_layer.loadNamedStyle(qml_file)
            # Set opacity:
            #active_raster_layer.setLayerTransparency(40)    # %    this method seems only be available for vector layer
            active_raster_layer.renderer().setOpacity(0.6)  # 0.6 = 40% transparency
            
            if self.dock.check_excl_zeroes.isChecked():
                self.out("setting zeroes to 100% transparency")
                '''
                # Set 0 values to NODATA (= transparent):
                provider = active_raster_layer.dataProvider()
                provider.setNoDataValue(1, 0)    # first one is referred to band number 
                # -> is working
                '''
                # better, conserves 0 value:
                tr = QgsRasterTransparency()
                tr.initializeTransparentPixelList(0)
                active_raster_layer.renderer().setRasterTransparency(tr)
            
            if hasattr(active_raster_layer, "setCacheImage"):
                active_raster_layer.setCacheImage(None)
            
            active_raster_layer.triggerRepaint()    # muss aufgerufen werden, Layer bleibt sonst schwarzweiß
            # Das ist für die QML-Farbskala im Layerfenster:
            #self.iface.legendInterface().refreshLayerSymbology(active_raster_layer)    # QGIS 2
            self.iface.layerTreeView().refreshLayerSymbology(active_raster_layer.id())
        # qml
        
        
        if write:
            yyyymmddHHMM = time.strftime("%Y%m%d_%H%M")
            new_file = path.join(self._model.data_dir, yyyymmddHHMM + '.qgs')
            self.out("write: {}".format(new_file))
            project.write(new_file)
        
        
        
        #self.dock.button_box.button(QDialogButtonBox.Apply).setEnabled(False)
        #self.dock.btn_action.setEnabled(False)    # action was performed
        # -> better not, when user modify some options and wants to perform the same work again...
        
        # Close dock and perform work:
        self.dock.close()
        # !!! Close dock at the end, otherwise the isVisible() method on elements will return
        # False because the visibility of the parent element is also relevant !!!
        
        self.dock.tabWidget.setCurrentIndex(1)    # show statistics tab when open dock next
        
        """
        Disabled because it is not working yet in QGIS 3
        """
        self.out("In QGIS 3 the print layout doesn't work anymore. If you can do it ...", False)
        #self._load_print_layout()
        self.out("*** Whole process finished! ***")
        
        self.iface.messageBar().pushMessage("Success", "Layer loaded", level=Qgis.Success, duration=3)
        
    
    
    def write_new_storage_location(self):
        data_dir_def_file = self._model.data_dir_def_file
        path_to_save = self.dock.storage_dir
        
        with open(data_dir_def_file, 'w') as f:
            f.write(path_to_save)
        
        self.out("write_new_storage_location()")
        print("  saved path '{}' to file '{}'".format(path_to_save, data_dir_def_file))
        
        self.dock.btn_save.setEnabled(False)    # show user, that action was performed
        
        # update new data path:
        self._model.data_dir = path_to_save
        
        #self.dock.tabWidget.addTab(0, 'tab_action')
        self.dock.tabWidget.setTabEnabled(0, True)
        # -> tab evtl. restored
        
    
    
    def _load_print_layout(self):
        
        window_title = "Print"    # gleich mit aktivem PrintComposer-Fensternamen geprüft
        
        """
        Vermeiden, immer wieder einen neuen PrintComposer zu erzeugen.
        Sonst kommen immer mehr Composer in die Liste - auch unter gleichem Namen.
        """
        composition = None
        
        layoutmanager = QgsLayoutManager()
        
        #for composer in self.iface.activeComposers():    # QGIS 2
        for composer in layoutmanager.layouts():
            #if composer.isActiveWindow():    # egal
            if composer.composerWindow().windowTitle() == window_title:
                self.out("active composer found")
                composition = composer
                break
        
        
        if not composition:
            self.out("no composer found; creating one...")
            
            # Load the template into the composer
            # QGIS 2:
            #active_composer = self.iface.createNewComposer(window_title)    #createNewComposer()
            #active_composer = QgsComposition(QgsProject.instance())
            
            project = QgsProject()
            #composition = QgsLayout(project)
            composition = QgsPrintLayout(project)
            composition.initializeDefaults()    # notwendig?
            
            q_xmldoc = self._model.create_qdocument_from_print_template_content()
            
            # load layout from template and add to Layout Manager
            #composition.composition().loadFromTemplate(q_xmldoc)    # QGIS 2
            composition.loadFromTemplate(q_xmldoc, QgsReadWriteContext())
            project.layoutManager().addLayout(composition)
        # if
        
        #composition = active_composer.composition()    # nicht verwechseln mit 'active_composer'
        
        #logo_item = composition.getComposerItemById('logo')    # QGIS 2
        logo_item = composition.itemById('logo')
        #self.out(">>> logo_item: '{}'".format(logo_item))    # gibt nur die Objekt-ID aus
        logo_image = path.join(self._model.layout_dir, "logo.tif")
        self.out("Logo: {}".format(logo_image))
        if path.exists(logo_image):
            #logo_item.setPictureFile(logo_image)    # QGIS 2
            logo_item.setPicturePath(logo_image)
        else:
            self.out("  ERROR: nicht gefunden!", False)
        
        
        """
        Hier versuche ich ein für die Überschrift mit einer ID ('headline') versehenes
        QgsLabel aus dem Template ausfindig zu machen. Ich mache das hier sehr kompliziert,
        es gibt bestimmt einen einfacheren Weg.
        Folgendes hat NICHT funktioniert:
        map_item = active_composer.getComposerItemById('headline')
        print(active_composer.items())
        liefert: [<PyQt4.QtGui.QGraphicsRectItem object at 0x124c60e8>, <qgis._core.QgsComposerLabel object at 0x124c65a8>, ... ]
        """
        found = False
        prod_id = self._ascii_converter.prod_id
        
        for item in list(composition.items()):
            #if type(item) != QgsComposerLabel:
            #    continue
            
            # QgsComposerLabel:
            composerLabel = item    # rename
            #print("->", composerLabel.id() )
            
            # Ein QgsComposerLabel wurde im Template mit der ID 'headline' versehen.
            if composerLabel.id() == "headline":
                title    = self._model.title(self._ascii_converter.date, prod_id)
                subtitle = ""
                if prod_id != 'RW':
                    subtitle = "{}-Produkt (Basis: RW)".format(prod_id)
                
                if prod_id != 'SF':
                    if subtitle:
                        subtitle += '\n'
                    subtitle += "für <Ort>"
                
                composerLabel.setText( title + "\n" + subtitle )
                #active_composer.refreshItems()    # wohl früher gebraucht
                found = True
                break
        # for
        
        # Hinweis, dass ggf. Template überarbeitet werden muss:
        if not found:
            self.out("no element with id 'headline' found!", False)
        
        
        
        legend = None
        
        """ TODO: schneller, wenn wir der Legende im Template eine ID vergeben:
        legend = newcomp.composition().getComposerItemById('legend')
        """
        for item in list(composition.items()):
            #if isinstance(item, QgsComposerLegend):
            if item.id() == "legend":
                #print i
                #print i.model()
                legend = item
                break
                """
                for i in xrange(legend.modelV2().rowCount()):
                    posteleg = legend.modelV2().index(i, 0)
                    print(posteleg.data())
                """
        # for
        
        if not legend:
            self.out("Legende konnte nicht erzeugt werden!", False)
            return
        
        
        #legend.setTitle("mm")    # Rasterlayer hat schon diesen Namen
        
        #
        # Layer für die Legende ausfindig machen
        #
        
        # You would just need to make sure your layer has a name you can distinguish from others. Instead of:
        # Vorherige Version:
        #active_raster_layer = self.iface.activeLayer()
        radolan_layer_name = "mm"
        # do:
        l_layer = QgsProject.instance().mapLayersByName(radolan_layer_name)
        active_raster_layer = l_layer[0] if l_layer  else None
        
        if not active_raster_layer:
            self.out("legend: kein Layer mit Namen '{}' gefunden! (active_raster_layer=None)".format(radolan_layer_name), False)
            return
        
        #print("Legend active_raster_layer id:",   active_raster_layer.id())     # ok
        #print("Legend active_raster_layer name:", active_raster_layer.name())   # ok
        #legend.model().setLayerSet([layer.id() for layer in layers])
        #legend.model().setLayerSet([active_raster_layer.id()])    # bringt nichts
        # DAS ist es! Dies fügt zumindest erstmal das interessierende Rasterlayer hinzu:
        #legend.modelV2().rootGroup().addLayer(active_raster_layer)
        #legend.updateLegend()
        
        l = QgsProject.instance().layoutManager()
        
        for layout in l.printLayouts():
            legend = QgsLayoutItemLegend(layout)
            root = QgsLayerTree()
            #for lyr in self.iface.mapCanvas().layers():
            #    if lyr.name() != 'Unwanted layer':
            #        root.addLayer(lyr)
            root.addLayer(active_raster_layer)
            legend.model().setRootGroup(root)
            layout.addItem(legend)
        
        """ By default the newly created composer items have zero position (top left corner of the page) and zero size.
        The position and size are always measured in millimeters.
        # set label 1cm from the top and 2cm from the left of the page
        composerLabel.setItemPosition(20, 10)
        # set both label’s position and size (width 10cm, height 3cm)
        composerLabel.setItemPosition(20, 10, 100, 30)
        A frame is drawn around each item by default. How to remove the frame:
        composerLabel.setFrame(False)
        """
        #print(active_composer.rect().width(), active_composer.rect().height())                   # 1054 911
        #print(self.iface.mapCanvas().size().width(), self.iface.mapCanvas().size().height())     # 1517 535
        # "Leinwandgröße": habe keine vernünftigen Werte oben ermittelt (vielleicht Pixel; ich brauche mm).
        # selbst, mittels Mauszeiger ermittelt:
        width  = 210
        height = 297
        # Rand neben der Legende:
        dw = 10
        dh = 14
        
        """
        Funktioniert mit QGIS 3 nicht mehr und ich bekomme es auch noch nicht hin. 28.07.2019
        """
        
        #legendSize = legend.paintAndDetermineSize(None)
        #legend.setItemPosition(width - legendSize.width() - dw,  height - legendSize.height() - dh)
        #legend.setItemPosition(width - legend.width() - dw,  height - legend.height() - dh)
        
        
        '''
        Ggf. alternativ noch das hier probieren:
        https://gis.stackexchange.com/questions/269857/pyqgis-update-refresh-composer-legend
        # legendLyr.renderer() nicht für Vektor-Layer
        if legendLyr.type() != QgsMapLayer.RasterLayer:
            continue
        '''
        
        
        
        # Also note that active_composer.composerWindow() has a hide() and show()
        #active_composer.composerWindow().hide()    # funktioniert
        
        """
        Rückgabe eines Signals oder Callables(?) notwendig, wenn ich diese Methode
        testweise über den Template-Button aufrufe.
        
        That calls test which returns None.
        .connect is expecting a variable of some type, instead of None.
        Either try just test instead of test(), or read up on what it expects and use that.
        """
        return pyqtSignal()    # einfach so etwas genommen. Damit funktioniert es.
        # http://pyqt.sourceforge.net/Docs/PyQt4/new_style_signals_slots.html
    
    
    
    def ask_user_load_template_project(self):
        reply = QMessageBox.question(self.iface.mainWindow(), 'Continue?',
                 'Do you want to load template project?', QMessageBox.Yes, QMessageBox.No)
        
        if reply == QMessageBox.No:
            return
        
        
        self.load_project()
        
        # Show hint:
        msg = "Please save project immediately with a new name,\nnot to overwrite the template!"
        QMessageBox.warning(self.iface.mainWindow(), 'Save as new project', msg)
        
        # trigger the Project Save As menu button.
        self.iface.mainWindow().findChild( QAction, 'mActionSaveProjectAs' ).trigger()
        
    
    
    def load_project(self):
        ''' Load a prepared template project.
        see https://docs.qgis.org/testing/en/docs/pyqgis_developer_cookbook/loadproject.html '''
        
        self.out("load_project()")
        
        project_file = self._model.default_project
        #print(project_file)
        # Get the project instance
        project = QgsProject.instance()
        # current project
        
        # Load another project:
        
        self.out("loading template project: {}".format(project_file))
        #project.read( QFileInfo(project_file) )    # funktioniert in QGIS3 nicht mehr
        project.read(project_file)
        
        return project
        
        